print("import: 'pymc'")
import pymc

print("import: 'pymc.backends'")
import pymc.backends

print("import: 'pymc.distributions'")
import pymc.distributions

print("import: 'pymc.gp'")
import pymc.gp

print("import: 'pymc.step_methods'")
import pymc.step_methods

print("import: 'pymc.tests'")
import pymc.tests

print("import: 'pymc.tuning'")
import pymc.tuning

print("import: 'pymc.variational'")
import pymc.variational

print("import: 'pymc.ode'")
import pymc.ode

